(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived_multiple.
Require Import Derived_env.


(*************************************************************************)
(* N-ary type functions *)
(*************************************************************************)

(* k1 .. kN -> k   :=   k1 -> ... -> kN -> k *)
Fixpoint arrsK ks k {struct ks} : knd :=
  match ks with
  | nil => k
  | k'::ks' => arrK k' (arrsK ks' k)
  end.

(* \a1 : k1 ... aN : kN. t   :=   \a1 : k1. ... . \aN : kN. t *)
Fixpoint lamsT ks t {struct ks} : typ :=
  match ks with
  | nil => t
  | k'::ks' => lamT k' (lamsT ks' t)
  end.

(* t t1 ... tN   :=   t t1 ... tN *)
Fixpoint appsT t ts {struct ts} : typ :=
  match ts with
  | nil => t
  | t'::ts' => (appsT (appT t t') ts')
  end.


(*************************************************************************)
(* Substitution *)
(*************************************************************************)

Lemma substAT_lams : forall a t' ks t,
  substAT a t' (lamsT ks t) = lamsT ks (substAT (shifts (length ks) a) t' t).
Proof.
  intros a t' ks t.
  generalize dependent a. generalize dependent t.
  induction ks; intros t a'; simpl lamsT; simpl substAT.
    Case "ks = nil".
      reflexivity.
    Case "ks = k::ks'".
      rewrite IHks. rewrite shift_shifts. rewrite shifts_shift. reflexivity.
Qed.

Corollary openAT_lams : forall t' ks t,
  openAT t' (lamsT ks t) = lamsT ks (substAT (length ks) t' t).
Proof.
  intros t' ks t. unfold openAT.
  rewrite substAT_lams. rewrite shifts_bnd. rewrite plus_zero. reflexivity.
Qed.


Lemma substAT_apps : forall a t' t ts,
  substAT a t' (appsT t ts) = appsT (substAT a t' t) (map (substAT a t') ts).
Proof.
  intros a t' t ts.
  generalize dependent t.
  induction ts; intros t; simpl appsT; simpl substAT.
    Case "ts = nil".
      reflexivity.
    Case "ts = t'::ts'".
      rewrite IHts. simpl substAT. reflexivity.
Qed.

Corollary openAT_apps : forall t' t ts,
  openAT t' (appsT t ts) = appsT (openAT t' t) (map (openAT t') ts).
Proof.
  intros t' t ts. unfold openAT.
  rewrite substAT_apps. unfold substAT. reflexivity.
Qed.

Lemma opensAT_apps : forall ts' t ts,
  opensAT ts' (appsT t ts) = appsT (opensAT ts' t) (map (opensAT ts') ts).
Proof.
  induction ts'; intros t ts; simpl opensAT.
    f_equal. simpl map. symmetry. apply map_id.
    rewrite IHts'. rewrite substAT_apps. f_equal. apply map_map.
Qed.


(*************************************************************************)
(* Local closure *)
(*************************************************************************)

Lemma lclosedT_lams : forall L ks t,
  (forall as', fresh L (length ks) as' -> lclosedT (opensAT (freeTs as') t)) ->
  lclosedT (lamsT ks t).
Proof.
  intros L ks. generalize dependent L.
  induction ks; intros L t Ht; simpl lamsT.
    replace t with (opensAT (freeTs nil) t).
      apply Ht. simpl fresh. trivial.
      unfold freeTs. simpl map. simpl opensAT. reflexivity.
    apply lclosedT_lam with (L := L). intros a0 Fr.
      rewrite openAT_lams. apply IHks with (L := L \u {{a0}}). intros as' Frs.
        replace (length ks) with (length (freeTs as')).
          rewrite <- opensAT_cons.
            rewrite <- freeTs_cons. apply Ht. simpl fresh. split. apply Fr. apply Frs.
            apply lclosedT_var.
            apply lclosedTs_vars. rewrite freeTs_length. symmetry. eapply fresh_length. apply Frs.
Qed.

Lemma lclosedT_apps : forall t ts,
  lclosedT t ->
  lclosedTs ts ->
  lclosedT (appsT t ts).
Proof.
  intros t ts. generalize dependent t.
  induction ts; intros t Ht Hts; simpl appsT.
    apply Ht.
    apply IHts.
      apply lclosedT_app.
        apply Ht.
        eapply lclosedTs_head. apply Hts.
      eapply lclosedTs_tail. apply Hts.
Qed.


(*************************************************************************)
(* Typing *)
(*************************************************************************)

Lemma ofT_lams : forall L D ks t k,
  (forall as', fresh L (length ks) as' -> ofT (D & as'~~ks) (opensAT (freeTs as') t) k) ->
  ofT D (lamsT ks t) (arrsK ks k).
Proof.
  intros L D ks t k.
  generalize dependent k. generalize dependent t.
  generalize dependent D. generalize dependent L.
  induction ks; intros L D t k Ht; simpl lamsT in *; simpl arrsK in *.
    Case "ks = nil".
      assert (Ht' : ofT (D & nil~~nil) (opensAT (freeTs nil) t) k).
        apply Ht. simpl fresh. trivial.
      apply Ht'.
    Case "ks = k' :: ks'".
      apply ofT_lam with (L := L).
        intros a' Fr. rewrite openAT_lams. apply IHks with (L := L \u {{a'}}).
          intros as' Frs.
            rewrite concat_assoc. rewrite <- concat_multi_cons.
            replace (length ks) with (length as').
              rewrite <- opensAT_cons_vars. apply Ht.
                simpl fresh. split. apply Fr. apply Frs.
              symmetry. eapply fresh_length. apply Frs.
Qed.

Lemma eqT_lams : forall L ks t t',
  (forall as', fresh L (length ks) as' -> eqT (opensAT (freeTs as') t) (opensAT (freeTs as') t')) ->
  eqT (lamsT ks t) (lamsT ks t').
Proof.
  intros L ks. generalize dependent L.
  induction ks; intros L t t' Ht; simpl lamsT.
    Case "ks = nil".
      replace t with (opensAT (freeTs nil) t). replace t' with (opensAT (freeTs nil) t').
        apply Ht. unfold length, fresh. trivial.
        unfold freeTs, opensAT. simpl. reflexivity.
        unfold freeTs, opensAT. simpl. reflexivity.
    Case "ts = t::ts".
      eapply eqT_lam. intros a0 Fr.
      rewrite openAT_lams. rewrite openAT_lams.
      apply IHks with (L := L \u {{a0}}). intros as' Frs.
      replace (length ks) with (length (freeTs as')).
      rewrite <- opensAT_cons. rewrite <- opensAT_cons.
        rewrite <- freeTs_cons. apply Ht. simpl fresh. split. apply Fr. apply Frs.
        apply lclosedT_var. apply lclosedTs_vars.
        apply lclosedT_var. apply lclosedTs_vars.
        rewrite freeTs_length. symmetry. eapply fresh_length. apply Frs.
Qed.

Lemma ofT_lams_eta : forall ks t,
  lclosedT t ->
  eqT (lamsT ks (appsT t (bndTs (length ks)))) t.
Proof.
  induction ks; intros t Ht; simpl lamsT; simpl appsT.
    Case "ks = nil".
      apply eqT_refl. apply Ht.
    Case "ks = k::ks'".
      fold (bnds (length ks)). fold (bndTs (length ks)).
      apply eqT_trans with (t' := lamT a (appT t 0)).
        apply eqT_lam with (L := {}).
          intros a0 Fr.
          rewrite openAT_lams. rewrite substAT_apps.
          unfold openAT. simpl substAT. destruct (length ks === length ks).
            rewrite substAT_bnd_bndTs. rewrite substAT_lclosed. rewrite substAT_lclosed.
              apply IHks. apply lclosedT_app. apply Ht. apply lclosedT_var.
              apply Ht. apply Ht.
              omega.
            destruct n. reflexivity.
        apply eqT_lam_eta. apply Ht.
Qed.

Lemma ofT_apps : forall D t ts ks k,
  ofT D t (arrsK ks k) ->
  length ts = length ks ->
  (forall t' k', In (t', k') (combine ts ks) -> ofT D t' k') ->
  ofT D (appsT t ts) k.
Proof.
  intros D t ts ks k.
  generalize dependent ks. generalize dependent k. generalize dependent t.
  induction ts; intros t k ks Ht Len Hts; simpl appsT in *.
    Case "ts = nil".
      destruct ks; simpl arrsK in *; simpl length in Len.
        SCase "ks = nil".
          apply Ht.
        SCase "ks = k'::ks'".
          inversion Len.
    Case "ts = t'::ts'".
      destruct ks; simpl arrsK in *; simpl length in Len.
        SCase "ks = nil".
          inversion Len.
        SCase "ks = k'::ks'".
          inversion Len.
          apply IHts with (ks := ks).
            eapply ofT_app.
              apply Ht.
              apply Hts. simpl combine. apply in_eq.
            apply H0.
            intros t' k' Elem. apply Hts. simpl combine. apply in_cons. apply Elem.
Qed.

Lemma eqT_apps : forall t ts t' ts',
  eqT t t' ->
  length ts = length ts' ->
  (forall ti ti', In (ti, ti') (combine ts ts') -> eqT ti ti') ->
  eqT (appsT t ts) (appsT t' ts').
Proof.
  intros t ts. generalize dependent t.
  induction ts; intros t t' ts' Ht Len Hts.
    Case "ts = nil".
      destruct ts'; simpl appsT.
        SCase "ts' = nil".
          apply Ht.
        SCase "ts' <> nil".
          simpl length in Len. inversion Len.
    Case "ts = t::ts".
      destruct ts'; simpl appsT.
        SCase "ts' = nil".
          simpl length in Len. inversion Len.
        SCase "ts' = t'::ts'".
          apply IHts.
            apply eqT_app.
              apply Ht.
              apply Hts. simpl combine. apply in_eq.
            simpl in Len. inversion Len. reflexivity.
            intros ti ti' Elem. apply Hts. simpl combine. apply in_cons. apply Elem.
Qed.

Lemma eqT_apps_beta : forall ks t ts,
  lclosedT t ->
  length ts = length ks ->
  lclosedTs ts ->
  eqT (appsT (lamsT ks t) ts) (opensAT ts t).
Proof.
  induction ks; intros t ts Ht Len Hts.
    Case "ks = nil".
      destruct ts; simpl appsT; simpl lamsT.
        SCase "ts = nil".
          unfold opensAT. apply eqT_refl. apply Ht.
        SCase "ts <> nil".
          simpl length in Len. inversion Len.
    Case "ts = t::ts".
      destruct ts; simpl appsT; simpl lamsT.
        SCase "ts = nil".
          simpl length in Len. inversion Len.
        SCase "ts = t'::ts'".
          apply eqT_trans with (t' := appsT (lamsT ks (substAT (length ks) t0 t)) ts).
            apply eqT_apps.
              rewrite <- openAT_lams. apply eqT_app_beta.
                change ( lclosedT (lamsT (a::ks) t)).
                apply lclosedT_lams with (L := {}). intros as' Frs. rewrite opensAT_lclosed; apply Ht.
              eapply lclosedTs_head. apply Hts.
              reflexivity.
              intros ti ti' Elem.
                replace ti' with ti.
                  apply eqT_refl. eapply lclosedTs_rest. apply Hts. eapply in_combine_l. apply Elem.
                  eapply combine_refl. apply Elem.
            rewrite opensAT_cons. rewrite substAT_lclosed. rewrite substAT_lclosed.
              apply IHks.
                apply Ht.
                simpl length in Len. inversion Len. reflexivity.
                eapply lclosedTs_tail. apply Hts.
              apply Ht. apply Ht.
              eapply lclosedTs_head. apply Hts.
              eapply lclosedTs_tail. apply Hts.
Qed.
